"""Miscellaneous utility functions and classes"""

from __future__ import annotations

__all__ = (
    'StatusEnum', 'LogEnum', 'Log', 'Job', 'has_path_seps', 'import_object',
    'terminate', 'get_env_log_level'
)

import importlib
import os
import signal
import time
from dataclasses import dataclass, field
from enum import auto, Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from datetime import datetime
    from typing import Any, Optional
    from uuid import UUID

    from .type import Terminatable


class StatusEnum(Enum):
    """Job statuses"""
    queued = auto()
    """Job was created, but hasn't been started by a worker yet"""
    downloading = auto()
    """Job is currently running"""
    finished = auto()
    """Job completed successfully"""
    error = auto()
    """Job failed"""
    cancelled = auto()
    """Job was cancelled by the user"""
    timeout = auto()
    """Celery task running the job timed out"""


class LogEnum(Enum):
    """YoutubeDL log levels"""
    debug = auto()
    warning = auto()
    error = auto()


@dataclass
class Log:
    level: LogEnum
    message: str
    timestamp: datetime

    def get_json(self) -> dict[str, Any]:
        return {
            'level': self.level.name,
            'message': self.message,
            'timestamp': self.timestamp.isoformat()
        }


@dataclass
class Progress:
    progress: dict[str, Any]
    timestamp: datetime
    filename: str = field(init=False)

    def __post_init__(self) -> None:
        self.filename = self.progress['filename']

    def get_json(self) -> dict[str, Any]:
        return self.progress


@dataclass
class Job:
    id: UUID
    status: StatusEnum
    urls: Sequence[str]
    ytdl_opts: Optional[Mapping[str, Any]] = None
    started: Optional[datetime] = None
    finished: Optional[datetime] = None
    logs: Sequence[Log] = tuple()
    error: Optional[str] = None
    progress: Sequence[Progress] = tuple()

    @staticmethod
    def _iso_or_none(datetime_: Optional[datetime]) -> Optional[str]:
        """Returns an ISO timestamp or None"""
        if datetime_ is not None:
            return datetime_.isoformat()
        else:
            return None

    def get_json(self) -> dict[str, Any]:
        return {
            'id': str(self.id),
            'status': self.status.name,
            'urls': tuple(self.urls),
            'ytdl_opts': self.ytdl_opts,
            'started': self._iso_or_none(self.started),
            'finished': self._iso_or_none(self.finished),
            'logs': tuple(
                log.get_json() for log in self.logs
            ),
            'error': self.error,
            'progress': tuple(
                progress.get_json() for progress in self.progress
            )
        }


def _kill(pid: int) -> None:
    """Manually kill the given PID by sending SIGKILL

    Raises `NotImplementedError` on Windows because it doesn't have
    SIGKILL
    """
    if hasattr(signal, 'SIGKILL'):
        os.kill(pid, signal.SIGKILL)
    else:
        raise NotImplementedError("SIGKILL isn't supported on Windows")


def _join(process: Terminatable, timeout: int) -> None:
    """Manual re-implementation of `Process.join()`

    This is used because `join()` randomly raises an exception when
    using `billiard.Process`
    (https://github.com/celery/billiard/issues/270)
    """
    for _ in range(timeout):
        time.sleep(1)
        if not process.is_alive():
            return


def terminate(process: Terminatable, timeout: int = 10) -> bool:
    """Terminate the given process

    First, SIGTERM is sent to the process. If the process doesn't exit
    within `timeout` seconds, SIGKILL is sent.

    If SIGKILL is unable to kill the process within 5 seconds,
    `ChildProcessError` is raised

    Returns `True` if SIGTERM successfully terminated the process, or
    `False` if SIGKILL was needed

    Since the function waits on both SIGTERM and SIGKILL, the maximum
    amount of time that the function can wait is `timeout` + 5 seconds
    """
    if not process.is_alive():
        return True

    process.terminate()
    _join(process, timeout)

    if process.is_alive():
        skip_wait = False

        # Unlike `multiprocessing.Process`, `billiard.Process` doesn't
        # have a `kill()` method, so we have to send SIGKILL manually
        if hasattr(process, 'kill'):
            # Ignore typing error about `kill()` not being defined
            # because we just checked it
            # (https://github.com/python/mypy/issues/1424)
            process.kill()  # type: ignore[attr-defined]
        else:
            try:
                # Ignore typing error about `process.pid` being the
                # wrong type. `process.pid` always returns an int after
                # the process is started
                _kill(process.pid)  # type: ignore[arg-type]
            except NotImplementedError:
                skip_wait = True

        if not skip_wait:
            _join(process, 5)

        if process.is_alive():
            raise ChildProcessError(
                'Timed out while waiting for the process to die'
            )
        else:
            return False
    else:
        return True


def import_object(module: str, attr: str) -> Any:
    """Import the given module, and return the given attribute from it"""
    module_imported = importlib.import_module(module)
    attr_value = getattr(module_imported, attr)
    return attr_value


def has_path_seps(str_: str) -> bool:
    """Returns `True` if the string contains path separators"""
    has_sep = os.sep in str_
    has_altsep = os.altsep is not None and os.altsep in str_
    return has_sep or has_altsep


def get_env_log_level() -> Optional[str]:
    """Get the log level from the env-var `$YTDL_LOG_LEVEL`

    This allows configuring the log level for both Flask and Celery with
    a single env-var.

    Returns `None` if the env-var isn't set. In this case, the logging
    shouldn't be configured.
    """
    log_level = os.environ.get('YTDL_LOG_LEVEL', None)
    if log_level:
        return log_level.upper()
    else:
        return None



"""Custom types for type-checking"""

from __future__ import annotations

__all__ = (
    'JSONableTypes', 'Terminatable', 'ParseMetadataDict',
    'ReplaceInMetadataDict'
)

from typing import Protocol, TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any, Optional, Union

    JSONableTypes = Union[
        type[dict], type[list], type[tuple[Any, ...]],
        type[str], type[int], type[float], type[bool], type[None]
    ]
    """Types that can be serialized to JSON

    Used by some subclasses of `error.http_error.HTTPError`
    """


class Terminatable(Protocol):
    """Protocol that represents a process that can be terminated

    Objects that match this protocol can be used as an argument with
    `util.terminate()`

    Example: `multiprocessing.Process`

    A protocol is used so that it will also accept alternative
    process objects such as `billiard.Process`
    """
    @property
    def pid(self) -> Optional[int]:
        ...

    def is_alive(self) -> bool:
        ...

    def join(self, timeout: Optional[int] = None) -> Any:
        ...

    def terminate(self) -> Any:
        ...


# This uses the old syntax because 'from' isn't a valid name for a class
# var
ParseMetadataDict = TypedDict(
    'ParseMetadataDict', {
        'from': str,
        'to': str
    }
)
"""Type that represents a single value of the parse_metadata custom_opt

Used by `custom_opt._generate_metadata_actions()`
"""


class ReplaceInMetadataDict(TypedDict):
    """Type that represents a single value of the replace_in_metadata
    custom_opt

    Used by `custom_opt._generate_metadata_actions()`
    """
    fields: Iterable[str]
    regex: str
    replace: str



"""Module for creating the Flask app"""

from __future__ import annotations

__all__ = ('create_app', 'with_logging')

from typing import TYPE_CHECKING
import logging

from flask import Flask

from . import celery, flask_config, health, job, json, meta
from .config import Config
from .database import db
from .util import get_env_log_level

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any, Optional


def _init_db(app: Flask) -> None:
    """Initialize the SQLAlchemy database for the given Flask app

    Raises `RuntimeError` if the database dialect isn't PostgreSQL.
    """
    # Disable deprecation warning about this option
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

    db.init_app(app)

    with app.app_context():
        dialect = db.session.bind.dialect.name
    if dialect != 'postgresql':
        raise RuntimeError(
            f'SQL database dialect must be PostgreSQL, not {dialect}'
        )

    # Create the tables if they don't exist
    db.create_all(app=app)


def create_app(test_config: Optional[Mapping[str, Any]] = None) -> Flask:
    """Create and return the Flask app

    `test_config` is an optional dict that will be used to load the
    `config.Config` class. Intended for testing.
    """
    app = Flask(__name__)

    if test_config is None:
        config = Config(logger=app.logger)
    else:
        app.config['TESTING'] = True
        config = Config(
            logger=app.logger, config_dict=test_config, load_env_vars=False
        )

    app.config['YTDL_CONFIG'] = config
    app.config.from_mapping(config.flask_config)

    app.json_encoder = json.JSONEncoder
    app.register_blueprint(json.bp)
    app.register_blueprint(job.bp, url_prefix='/jobs')
    app.register_blueprint(flask_config.bp, url_prefix='/config')
    app.register_blueprint(meta.bp, url_prefix='/meta')
    app.register_blueprint(health.bp, url_prefix='/health')

    app.logger.debug(
        'Initializing database: %s', config.database_uri
    )
    _init_db(app)

    app.logger.debug('Loading Celery config: %s', config.celery_config)
    celery.init_celery(app, config.celery_config)

    return app


def with_logging() -> Flask:
    """Set up logging, and create the Flask app

    The log level is set via the `$YTDL_LOG_LEVEL` env-var. If the var
    isn't set, logging won't be set up, which means that the level will
    default to 'WARNING'.

    This can be used as the entrypoint when running ytdl-server through
    a WSGI server (such as Gunicorn).
    """
    log_level = get_env_log_level()
    if log_level is not None:
        log_level = log_level.upper()
        logging.basicConfig(level=log_level)

    return create_app()



"""SQLAlchemy database and models"""

from __future__ import annotations

__all__ = ('db', 'Job', 'Log', 'Progress')

from typing import TYPE_CHECKING

import sqlalchemy as sql
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.dialects.postgresql import JSONB, UUID

from .util import LogEnum, StatusEnum

if TYPE_CHECKING:
    from flask_sqlalchemy.model import DefaultMeta

# Allow using `db` as a parent class.
# https://github.com/dropbox/sqlalchemy-stubs/issues/76#issuecomment-595839159
db: DefaultMeta = SQLAlchemy()


# Ignore typing error about `db.Model` not being subclassable because
# it's `Any`
class Job(db.Model):  # type: ignore[misc]
    __tablename__ = 'ytdl_server_jobs'

    id = sql.Column(
        UUID(as_uuid=True), primary_key=True,
        server_default=sql.text('gen_random_uuid()'), comment='Job ID'
    )

    status = sql.Column(
        sql.Enum(StatusEnum), nullable=False, comment='Job status'
    )
    error = sql.Column(sql.UnicodeText, nullable=True, comment='Error message')

    urls = sql.Column(
        sql.ARRAY(sql.UnicodeText, dimensions=1, as_tuple=True),
        nullable=False, comment='Video URLs to download'
    )
    ytdl_opts = sql.Column(JSONB, nullable=True, comment='youtube-dl options')

    started = sql.Column(
        sql.TIMESTAMP(timezone=True), nullable=True, comment='Time job started'
    )
    finished = sql.Column(
        sql.TIMESTAMP(timezone=True), nullable=True,
        comment='Time job finished'
    )


class Log(db.Model):  # type: ignore[misc]
    __tablename__ = 'ytdl_server_logs'
    __table_args__ = (
        sql.Index(
            'ytdl_server_logs_job_id_timestamp_idx', 'job_id', 'timestamp'
        ),
    )

    id = sql.Column(
        UUID(as_uuid=True), primary_key=True,
        server_default=sql.text('gen_random_uuid()'), comment='Primary key'
    )
    job_id = sql.Column(
        UUID(as_uuid=True), sql.ForeignKey(
            Job.id, ondelete='CASCADE', onupdate='CASCADE'
        ), nullable=False, comment='Job ID'
    )
    level = sql.Column(sql.Enum(LogEnum), nullable=False, comment='Log level')
    message = sql.Column(
        sql.UnicodeText, nullable=False, comment='Log message'
    )
    timestamp = sql.Column(
        sql.TIMESTAMP(timezone=True), nullable=False,
        server_default=sql.text('CURRENT_TIMESTAMP'),
        comment='Time the message was logged'
    )


class Progress(db.Model):  # type: ignore[misc]
    __tablename__ = 'ytdl_server_progress'
    __table_args__ = (
        # Ensure that only one row exists for each file that's
        # downloaded instead of creating a new row for every progress
        # update
        sql.UniqueConstraint('job_id', 'filename'),
        sql.Index(
            'ytdl_server_progress_job_id_timestamp_idx', 'job_id', 'timestamp'
        )
    )

    id = sql.Column(
        UUID(as_uuid=True), primary_key=True,
        server_default=sql.text('gen_random_uuid()'), comment='Primary key'
    )
    job_id = sql.Column(
        UUID(as_uuid=True), sql.ForeignKey(
            Job.id, ondelete='CASCADE', onupdate='CASCADE'
        ), nullable=False, comment='Job ID'
    )
    filename = sql.Column(
        sql.UnicodeText, nullable=False,
        comment='Filename of the downloaded file'
    )
    progress = sql.Column(
        JSONB, nullable=False,
        comment='Object containing the download progress info'
    )
    timestamp = sql.Column(
        sql.TIMESTAMP(timezone=True), nullable=False,
        server_default=sql.text('CURRENT_TIMESTAMP'),
        comment='Time the progress info was last updated'
    )



"""Manage custom_opts and convert them to ytdl_opts"""

from __future__ import annotations

__all__ = ('CustomOpt', 'CUSTOM_OPTS', 'load_custom_opts', 'merge_custom_opts')

import re
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING

from .util import import_object

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Iterable, Mapping
    from typing import Any, Final, Optional, Union

    from .task import TaskArgs
    from .type import ParseMetadataDict, ReplaceInMetadataDict


def load_custom_opts(args: TaskArgs) -> dict[str, Any]:
    """Convert the custom_opts dict into a ytdl_opts dict that can be
    passed to `ytdl.Downloader`"""
    loaded_opts: list[dict[str, Any]] = []
    merged_opts: dict[str, Any] = {}

    for opt in CUSTOM_OPTS:
        loaded = opt.load(args)
        if loaded:
            loaded_opts.append(loaded)

    # Merge the loaded opts
    #
    # The postprocessors lists are combined together at the end so that
    # they don't overwrite each other
    postprocessors = []
    for loaded in loaded_opts:
        if 'postprocessors' in loaded:
            postprocessors += loaded.pop('postprocessors')
        merged_opts |= loaded
    if postprocessors:
        merged_opts['postprocessors'] = postprocessors

    return merged_opts


def merge_custom_opts(
    ytdl_opts: Mapping[str, Any], loaded_custom_opts: Mapping[str, Any]
) -> dict[str, Any]:
    """Merge the loaded custom_opts into ytdl_opts

    For conflicting keys, the value in loaded_custom_opts will be used

    If both mappings contain the 'postprocessors' list, the ytdl_opts
    list will be appended to the loaded_custom_opts dict instead of
    overwriting it.
    This prevents custom_opts from overwriting postprocessors completely
    """
    PP: Final = 'postprocessors'
    postprocessors: Optional[tuple[Any, ...]] = None

    # Only merge the postprocessor lists if they're both sequences so
    # that it doesn't break if the user sets it to a weird type
    if (
        PP in ytdl_opts and PP in loaded_custom_opts and
        isinstance(ytdl_opts[PP], Sequence) and
        isinstance(loaded_custom_opts[PP], Sequence)
    ):
        postprocessors = tuple(loaded_custom_opts[PP]) + tuple(ytdl_opts[PP])

    merged_opts = dict(ytdl_opts)
    merged_opts |= loaded_custom_opts
    if postprocessors is not None:
        merged_opts[PP] = postprocessors

    return merged_opts


@dataclass(frozen=True)
class CustomOpt:
    """Dataclass for handling custom_opts and merging them into
    ytdl_opts"""

    custom_opts: tuple[str, ...]
    """List of custom_opts that are handled by this instance

    `handler` will only be called if *any* of these custom_opts are
    given by the user and their value isn't `None`
    """
    handler: Callable[[TaskArgs], Optional[dict[str, Any]]]
    """The function that will be used to load the custom_opts

    The function should return a dict of ytdl_opts, which will be merged
    into the final ytdl_opts.
    It can also return `None` if no options should be added
    """

    def load(self, args: TaskArgs) -> Optional[dict[str, Any]]:
        """Load the custom_opts based on the given TaskArgs

        Returns `None` if the custom_opts within `args` does not contain
        any corresponding custom_opts
        """
        for opt in self.custom_opts:
            if opt in args.custom_opts and args.custom_opts[opt] is not None:
                # mypy incorrectly infers the type of callable objects
                # when using a frozen dataclass:
                # https://github.com/python/mypy/issues/7404
                return self.handler(args)  # type: ignore
        else:
            return None


def _load_daterange(args: TaskArgs) -> dict[str, Any]:
    """Use 'dateafter' and 'datebefore' to create a DateRange object,
    which is assigned to the 'daterange' ytdl_opt

    'daterange' can't be given directly as a ytdl_opt because it's not
    JSONable
    """
    DateRange = import_object(args.daterange_module, args.daterange_class)
    dateafter = args.custom_opts.get('dateafter', None)
    datebefore = args.custom_opts.get('datebefore', None)

    daterange = DateRange(dateafter, datebefore)
    return {
        'daterange': daterange
    }


def _load_pp_sponsorblock(args: TaskArgs) -> dict[str, Any]:
    """Add the `SponsorBlock` postprocessor

    Requires yt-dlp.

    The 'sponsorblock' dict is also used by the 'remove_chapters' and
    'addmetadata' custom_opts.
    """
    sponsorblock: dict[str, Any] = args.custom_opts['sponsorblock']

    mark: list[str] = sponsorblock.get('mark') or []
    remove: list[str] = sponsorblock.get('remove') or []
    api: Optional[str] = sponsorblock.get('api', None)

    # Merge the categories and remove duplicates.
    #
    # The 'SponsorBlock' PP doesn't differentiate between the two lists.
    # Removal is handled in the 'remove_chapters' custom_opt.
    categories = frozenset(mark) | frozenset(remove)

    pp = {
        'key': 'SponsorBlock',
        'categories': categories,
        'when': 'pre_process'
    }
    # Add the API only if given so that yt-dlp can fallback to its
    # default value
    if api is not None:
        pp['api'] = api

    return {
        'postprocessors': (pp,)
    }


def _generate_metadata_actions(
    Actions: Any, parse_metadata: Iterable[ParseMetadataDict],
    replace_in_metadata: Iterable[ReplaceInMetadataDict]
) -> Generator[
    Union[tuple[Any, str, str], tuple[Any, str, str, str]],
    None, None
]:
    """Parse the MetadataParser custom_opts, and yield a list of actions
    to be added to the postprocessor

    `Actions` should be the `MetadataParserPP.Actions` enum, which has
    the members `INTERPRET` and `REPLACE`.

    The postprocessor accepts a mixed list of `INTERPRET` actions
    (generated from the 'parse_metadata' custom_opt) and `REPLACE`
    actions (generated from the 'replace_in_metadata' custom_opt).

    Syntax of `INTERPRET` actions:
        (Actions.INTERPRET, 'FROM', 'TO')

    Syntax of `REPLACE` actions:
        (Actions.REPLACE, 'FIELD', 'REGEX', 'REPLACE')
    """
    for interpreter in parse_metadata:
        yield (Actions.INTERPRET, interpreter['from'], interpreter['to'])

    for replacer in replace_in_metadata:
        for field in replacer['fields']:
            yield (
                Actions.REPLACE, field, replacer['regex'], replacer['replace']
            )


def _load_pp_metadataparser(args: TaskArgs) -> Optional[dict[str, Any]]:
    """Add the `MetadataParser` postprocessor

    The postprocessor requires the `MetadataParserPP.Actions` enum,
    which we have to import from yt-dlp.
    """
    MetadataParserPP = import_object(args.metadata_module, args.metadata_class)
    parse_metadata: Optional[Sequence[ParseMetadataDict]] = (
        args.custom_opts.get('parse_metadata', ())
    )
    replace_in_metadata: Optional[Sequence[ReplaceInMetadataDict]] = (
        args.custom_opts.get('replace_in_metadata', ())
    )

    if parse_metadata is None:
        parse_metadata = ()
    if replace_in_metadata is None:
        replace_in_metadata = ()

    actions = tuple(_generate_metadata_actions(
        MetadataParserPP.Actions, parse_metadata, replace_in_metadata
    ))

    if actions:
        return {
            'postprocessors': ({
                'key': 'MetadataParser',
                'actions': actions,
                'when': 'pre_process'
            },)
        }
    else:
        return None


def _load_pp_metadatafromtitle(args: TaskArgs) -> dict[str, Any]:
    """Add the `MetadataFromTitle` postprocessor"""
    return {
        'postprocessors': ({
            'key': 'MetadataFromTitle',
            'titleformat': args.custom_opts['metafromtitle']
        },)
    }


def _load_pp_ffmpegextractaudio(args: TaskArgs) -> dict[str, Any]:
    """Add the `FFmpegExtractAudio` postprocessor"""
    extractaudio = args.custom_opts['extractaudio']
    loaded_opts: dict[str, Any] = {}

    # Automatically set the format so that it only downloads audio when
    # possible
    keepvideo = args.ytdl_opts.get('keepvideo', False)
    format_ = args.ytdl_opts.get('format', None)
    if not keepvideo and format_ is None:
        loaded_opts['format'] = 'bestaudio/best'

    loaded_opts['postprocessors'] = ({
        'key': 'FFmpegExtractAudio',
        'preferredcodec': extractaudio.get('audioformat', 'best'),
        'preferredquality': str(extractaudio.get('audioquality', '5')),
        'nopostoverwrites': extractaudio.get('nopostoverwrites', False)
    },)

    return loaded_opts


def _load_pp_ffmpegvideoconverter(args: TaskArgs) -> dict[str, Any]:
    """Add the `FFmpegVideoConvertor` postprocessor"""
    recodevideo: str = args.custom_opts['recodevideo']
    return {
        'postprocessors': ({
            'key': 'FFmpegVideoConvertor',
            # Typo is intentional
            'preferedformat': recodevideo
        },)
    }


def _load_pp_ffmpegvideoremuxer(args: TaskArgs) -> Optional[dict[str, Any]]:
    """Add the `FFmpegVideoRemuxer` postprocessor

    Requires yt-dlp.

    This option is ignored if the 'recodevideo' custom_opt is also
    given.
    """
    remuxvideo: str = args.custom_opts['remuxvideo']
    recodevideo: Optional[str] = args.custom_opts.get('recodevideo', None)

    if recodevideo is not None:
        args.logger.warning(
            'Conflicting custom_opts: recodevideo and remuxvideo. '
            'Ignoring remuxvideo.'
        )
        return None

    return {
        'postprocessors': ({
            'key': 'FFmpegVideoRemuxer',
            # Typo is intentional
            'preferedformat': remuxvideo
        },)
    }


def _load_pp_modifychapters(args: TaskArgs) -> Optional[dict[str, Any]]:
    """Add the `ModifyChapters` postprocessor

    Requires yt-dlp.

    The PP is used by the 'remove_chapters' and 'sponsorblock'
    custom_opts.
    """
    force_keyframes: Optional[bool] = None

    remove_chapters: Optional[dict[str, Any]] = (
        args.custom_opts.get('remove_chapters', None)
    )
    if remove_chapters:
        patterns: Optional[list[str]] = remove_chapters.get('patterns', None)
        ranges: Optional[list[list[Union[int, float]]]] = (
            remove_chapters.get('ranges', None)
        )
        force_keyframes = remove_chapters.get('force_keyframes', None)
    else:
        patterns = None
        ranges = None

    sponsorblock: Optional[dict[str, Any]] = (
        args.custom_opts.get('sponsorblock', None)
    )
    if sponsorblock:
        sponsorblock_remove: Optional[list[str]] = (
            sponsorblock.get('remove', None)
        )
        sponsorblock_template: Optional[str] = (
            sponsorblock.get('template', None)
        )
        if force_keyframes is None:
            force_keyframes = sponsorblock.get('force_keyframes', None)
    else:
        sponsorblock_remove = None
        sponsorblock_template = None

    if force_keyframes is None:
        force_keyframes = False

    # Compile all of the regexes
    if patterns:
        regex_patterns: Optional[tuple[re.Pattern[str], ...]] = tuple(
            re.compile(pattern) for pattern in patterns
        )
    else:
        regex_patterns = None

    pp: dict[str, Any] = {
        'key': 'ModifyChapters',
        'force_keyframes': force_keyframes
    }

    # Only return the PP if any args that do something were given.
    #
    # This will be `False` if only the 'sponsorblock' custom_opt was
    # used, and it doesn't have the 'remove' field.
    return_pp = False

    # Add the native remove_chapters args if given
    if regex_patterns:
        pp['remove_chapters_patterns'] = regex_patterns
        return_pp = True
    if ranges:
        pp['remove_ranges'] = ranges
        return_pp = True

    # Add the SponsorBlock args if given
    if sponsorblock_remove:
        pp['remove_sponsor_segments'] = frozenset(sponsorblock_remove)
        return_pp = True
    if sponsorblock_template:
        pp['sponsorblock_chapter_title'] = sponsorblock_template

    if return_pp:
        return {
            'postprocessors': (pp,)
        }
    else:
        return None


def _load_pp_ffmpegmetadata(args: TaskArgs) -> Optional[dict[str, Any]]:
    """Add the `FFmpegMetadata` postprocessor

    This is used by the 'addmetadata', 'addchapters', and
    'sponsorblock' custom_opts.

    The 'add_chapters' and 'add_metadata' kwargs are exclusive to
    yt-dlp, so we only add them if a yt-dlp custom_opt ('addchapters' or
    'sponsorblock') was given.
    When using yt-dlp, they both default to `True` if not given.
    """
    addmetadata: Optional[bool] = args.custom_opts.get('addmetadata', None)
    addchapters: Optional[bool] = args.custom_opts.get('addchapters', None)
    sponsorblock: Optional[dict[str, Any]] = (
        args.custom_opts.get('sponsorblock', None)
    )
    if sponsorblock and 'mark' in sponsorblock:
        sponsorblock_mark: Optional[list[str]] = sponsorblock['mark']
    else:
        sponsorblock_mark = None

    if addchapters is None and not sponsorblock_mark:
        # Don't add the extra kwargs when the yt-dlp custom_opts aren't
        # given in order to maintain backwards-compatibility with
        # youtube-dl
        if addmetadata:
            return {
                'postprocessors': ({
                    'key': 'FFmpegMetadata'
                },)
            }
        else:
            return None
    else:
        # 'addchapters' and/or 'sponsorblock.mark' was given.
        # Requires yt-dlp.
        if addchapters is None:
            addchapters = bool(addmetadata or sponsorblock_mark)
        if addmetadata is None:
            addmetadata = False

        if addmetadata or addchapters:
            return {
                'postprocessors': ({
                    'key': 'FFmpegMetadata',
                    'add_chapters': addchapters,
                    'add_metadata': addmetadata
                },)
            }
        else:
            # Don't add the PP if both kwargs are False, since it won't
            # do anything.
            return None


def _load_pp_ffmpegsubtitlesconverter(args: TaskArgs) -> dict[str, Any]:
    """Add the `FFmpegSubtitlesConvertor` postprocessor"""
    return {
        'postprocessors': ({
            'key': 'FFmpegSubtitlesConvertor',
            'format': args.custom_opts['convertsubtitles']
        },)
    }


def _load_pp_ffmpegthumbnailsconverter(args: TaskArgs) -> dict[str, Any]:
    """Add the `FFmpegThumbnailsConvertor` postprocessor

    Requires yt-dlp.
    """
    return {
        'postprocessors': ({
            'key': 'FFmpegThumbnailsConvertor',
            'format': args.custom_opts['convertthumbnails'],
            'when': 'before_dl'
        },)
    }


def _load_pp_ffmpegembedsubtitle(args: TaskArgs) -> Optional[dict[str, Any]]:
    """Add the `FFmpegEmbedSubtitle` postprocessor"""
    if args.custom_opts['embedsubtitles']:
        return {
            'postprocessors': ({
                'key': 'FFmpegEmbedSubtitle'
            },)
        }
    else:
        return None


def _load_pp_embedthumbnail(args: TaskArgs) -> Optional[dict[str, Any]]:
    """Add the `EmbedThumbnail` postprocessor"""
    if args.custom_opts['embedthumbnail']:
        loaded_opts: dict[str, Any] = {}

        writethumbnail = args.ytdl_opts.get('writethumbnail', False)
        write_all_thumbnails = args.ytdl_opts.get(
            'write_all_thumbnails', False
        )

        # The thumbnail file will be deleted after it's embedded if this
        # is False
        already_have_thumbnail = bool(writethumbnail or write_all_thumbnails)
        if not already_have_thumbnail:
            loaded_opts['writethumbnail'] = True

        loaded_opts['postprocessors'] = ({
            'key': 'EmbedThumbnail',
            'already_have_thumbnail': already_have_thumbnail
        },)

        return loaded_opts
    else:
        return None


def _load_pp_ffmpegsplitchapters(args: TaskArgs) -> dict[str, Any]:
    """Add the `FFmpegSplitChapters` postprocessor"""
    split_chapters: dict[str, Any] = args.custom_opts['split_chapters']
    force_keyframes: Optional[bool] = (
        split_chapters.get('force_keyframes', None)
    )
    if force_keyframes is None:
        force_keyframes = False

    return {
        'postprocessors': ({
            'key': 'FFmpegSplitChapters',
            'force_keyframes': force_keyframes
        },)
    }


def _load_pp_xattrmetadata(args: TaskArgs) -> Optional[dict[str, Any]]:
    """Add the `XAttrMetadata` postprocessor"""
    if args.custom_opts['xattrs']:
        return {
            'postprocessors': ({
                'key': 'XAttrMetadata'
            },)
        }
    else:
        return None


def _load_final_ext(args: TaskArgs) -> Optional[dict[str, str]]:
    """Add the `final_ext` ytdl_opt if needed

    The final_ext ytdl_opt is used by yt-dlp to detect if the file is
    already downloaded. It's needed when a postprocessor changes the
    file extension.

    final_ext will be set to the value of the first existing custom_opt:
        recodevideo, remuxvideo, extractaudio.audioformat

    Based on the yt-dlp source:
        https://github.com/yt-dlp/yt-dlp/blob/df03de2c02192e43e5b51c8708619179a268b4cf/yt_dlp/__init__.py#L587-L591
    """
    # Don't override the existing value when it's manually set
    existing_final_ext: Optional[str] = args.ytdl_opts.get('final_ext', None)
    if existing_final_ext is not None:
        return None

    recodevideo: Optional[str] = args.custom_opts.get('recodevideo', None)
    remuxvideo: Optional[str] = args.custom_opts.get('remuxvideo', None)
    extractaudio: Optional[dict[str, Any]] = (
        args.custom_opts.get('extractaudio', None)
    )
    if extractaudio:
        audioformat: Optional[str] = extractaudio.get('audioformat', None)
    else:
        audioformat = None

    final_ext: Optional[str] = None

    # Only the first option that isn't `None` is checked so that we
    # don't check options that have been overridden.
    # For example, remuxvideo is ignored if recodevideo is set.
    if recodevideo is not None:
        # Only set final_ext if the option only contains alphanumeric
        # characters. This prevents it from being set if the option is
        # set to a value like 'aac>m4a/mov>mp4/mkv'.
        # yt-dlp handles this more accurately by checking if the value
        # is in the set of supported extensions, but we don't have
        # access to that information.
        if recodevideo.isalnum():
            final_ext = recodevideo
    elif remuxvideo is not None:
        if remuxvideo.isalnum():
            final_ext = remuxvideo
    elif audioformat is not None and audioformat != 'best':
        final_ext = audioformat

    if final_ext is not None:
        return {
            'final_ext': final_ext
        }
    else:
        return None


CUSTOM_OPTS: Final = (
    CustomOpt(('datebefore', 'dateafter'), _load_daterange),
    CustomOpt(('sponsorblock',), _load_pp_sponsorblock),
    CustomOpt(
        ('parse_metadata', 'replace_in_metadata'), _load_pp_metadataparser
    ),
    CustomOpt(('metafromtitle',), _load_pp_metadatafromtitle),
    CustomOpt(('convertsubtitles',), _load_pp_ffmpegsubtitlesconverter),
    CustomOpt(('convertthumbnails',), _load_pp_ffmpegthumbnailsconverter),
    CustomOpt(('extractaudio',), _load_pp_ffmpegextractaudio),
    CustomOpt(('remuxvideo',), _load_pp_ffmpegvideoremuxer),
    CustomOpt(('recodevideo',), _load_pp_ffmpegvideoconverter),
    CustomOpt(('embedsubtitles',), _load_pp_ffmpegembedsubtitle),
    CustomOpt(('remove_chapters', 'sponsorblock'), _load_pp_modifychapters),
    CustomOpt(
        ('addmetadata', 'addchapters', 'sponsorblock'), _load_pp_ffmpegmetadata
    ),
    CustomOpt(('embedthumbnail',), _load_pp_embedthumbnail),
    CustomOpt(('split_chapters',), _load_pp_ffmpegsplitchapters),
    CustomOpt(('xattrs',), _load_pp_xattrmetadata),
    CustomOpt(('recodevideo', 'remuxvideo', 'extractaudio'), _load_final_ext),
)
"""List of custom_opts that will be loaded

The order of the postprocessor custom_opts is important.
"""



"""Used by the `config` module to load the config options"""

from __future__ import annotations

__all__ = (
    'FlaskCeleryOption', 'Option', 'CELERY_OPTIONS', 'FLASK_OPTIONS', 'OPTIONS'
)

import os
from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Final, Optional

_DEFAULT: Final = object()


def _convert_env_var(value: str, type_: type, env_var: str) -> Any:
    """Convert the given env-var string to the correct type"""
    if type_ is str:
        return value
    elif type_ is bool:
        value = value.lower()
        if value in {'true', 't', '1', 'yes', 'y'}:
            return True
        elif value in {'false', 'f', '0', 'no', 'n'}:
            return False
        else:
            raise ValueError(
                f'The environment variable {env_var} has an invalid value:'
                f" {value!r}. Expected 'TRUE' or 'FALSE'"
            )
    elif type_ is list:
        return value.split(',')
    else:
        try:
            return type_(value)
        except Exception as e:
            raise ValueError(
                f'The environment variable {env_var} has an invalid value:'
                f' {value!r}. Expected an {type_.__name__}'
            ) from e


@dataclass(frozen=True)
class Option:
    """Individual config options used by `OPTIONS`"""

    name: str
    """Name of the option that will be used in the `config.Config` class"""
    type: type
    """Type of the option as found in the config file

    See `value_converter` if you want to change the type of the option
    after loading it
    """
    # The value is set to the correct type in `__post_init__()`
    config_key: tuple[str, ...] = None  # type: ignore[assignment]
    """Location of the option within the config file

    Multiple values represents a nested mapping. For example, a value
    of `('foo',)` represents the key 'foo' in the top-level mapping,
    while a value of `('foo', 'bar')` represents the key 'bar' within
    the mapping 'foo'

    Defaults to a tuple containing only `name`
    """
    env_var: Optional[str] = None
    """Environment var that the option will be loaded from if it exists

    The env var has priority over the value in the config file
    """
    default: Any = _DEFAULT
    """Default value of the option if it isn't found in the config file
    or as an env var"""
    default_factory: Optional[Callable[[], Any]] = None
    """Same as `default`, but the value is generated by calling this
    function

    Use this instead of `default` if the default value is mutable
    """
    required: bool = field(init=False)
    """Whether or not the option is required

    This is automatically set to `True` if either `default` or
    `default_factory` is defined
    """

    value_converter: Optional[Callable[[Any], Any]] = None
    """The value of the option will be passed through this function if
    defined

    For example, you can convert a list to a tuple by setting this to
    `tuple`

    See `convert_value()` for details

    Note: this function isn't used if the option is set to the default
    value
    """

    def __post_init__(self) -> None:
        if self.default is not _DEFAULT and self.default_factory is not None:
            raise ValueError(
                f'Conflicting options for {self.name!r}.'
                " Only one of 'default' or 'default_factory' can be defined"
            )

        required = self.default is _DEFAULT and self.default_factory is None
        object.__setattr__(self, 'required', required)

        if self.config_key is None:
            object.__setattr__(self, 'config_key', (self.name,))

    def get_default(self) -> Any:
        """Return the default value of the option

        Raises an error if the option is required
        """
        if self.required:
            raise RuntimeError(
                f'The option {self.name!r} does not have a default value'
            )
        if self.default is not _DEFAULT:
            return self.default
        else:
            # Ignore error about `None` not being callable, because
            # `default_factory` will never be `None` when this
            # else-statement runs
            return self.default_factory()  # type: ignore[misc]

    def convert_value(self, value: Any) -> Any:
        """Convert value of the option using `value_converter()`

        If `value_converter()` isn't defined, the value will be returned
        as-is
        """
        if self.value_converter is not None:
            return self.value_converter(value)
        else:
            return value

    def nested_in(
        self, dict_: Mapping[Any, Any], *, assert_type: bool = True
    ) -> bool:
        """Recursively check if the option is contained in the given
        config mapping

        If `assert_type` is `True`, an error will be raised if an
        intermediate mapping is the wrong type. Otherwise, it will just
        return `False`
        """
        for level in self.config_key:
            if not isinstance(dict_, Mapping):
                if assert_type:
                    raise TypeError(
                        f'The config option {".".join(self.config_key)!r} has '
                        'an invalid type. Expected a mapping, '
                        f'not a {type(dict_).__name__}'
                    )
                else:
                    return False
            if level not in dict_:
                return False
            dict_ = dict_[level]
        return True

    def nested_get(self, dict_: Mapping[Any, Any]) -> Any:
        """Recursively get the option from the config mapping"""
        for level in self.config_key:
            dict_ = dict_[level]
        # `dict_` is actually the value when the loop finishes
        return dict_

    def get_env(self, mapping: Optional[Mapping[str, str]] = None) -> Any:
        """Retrieve the option from the given env-var mapping

        The value will be converted from a str to the correct type.

        `mapping` defaults to `os.environ` when `None`.
        """
        if mapping is None:
            mapping = os.environ

        if self.env_var is None:
            raise RuntimeError(
                f'The option {self.name!r} does not have an env-var set'
            )

        value = mapping[self.env_var]
        return _convert_env_var(value, self.type, self.env_var)


@dataclass(frozen=True)
class FlaskCeleryOption:
    """Individual Flask-or-Celery config options used by `FLASK_OPTIONS`
    and `CELERY_OPTIONS`"""

    name: str
    """Name of the Flask/Celery option"""
    env_var: str
    """Environment var that the option will be loaded from if it exists"""
    type: type
    """Type of the option

    The option will be converted to this type
    """

    def get_env(self, mapping: Optional[Mapping[str, str]] = None) -> Any:
        """Retrieve the option from the given env-var mapping

        The value will be converted from a str to the correct type

        `mapping` defaults to `os.environ` when `None`.
        """
        if mapping is None:
            mapping = os.environ

        value = mapping[self.env_var]
        return _convert_env_var(value, self.type, self.env_var)


OPTIONS: Final = (
    Option(name='database_uri', env_var='YTDL_DATABASE_URI', type=str),
    Option(name='flask_config', type=dict, default_factory=dict),
    Option(name='celery_config', type=dict, default_factory=dict),
    # The download dir defaults to the working dir of the API (not the
    # worker) since it's passed as an arg to `task.download`.
    #
    # It's done this way because the API needs to know the download_dir
    # in order to check the value of `outtmpl`.
    Option(
        name='download_dir', env_var='YTDL_DOWNLOAD_DIR', type=str,
        default_factory=os.getcwd
    ),
    Option(
        name='force_download_dir', env_var='YTDL_FORCE_DOWNLOAD_DIR',
        type=bool, default=True
    ),
    Option(
        name='allow_user_opts', env_var='YTDL_ALLOW_USER_OPTS',
        type=bool, default=True
    ),
    Option(
        name='ytdl_default_opts', config_key=('ytdl_opts', 'default_opts'),
        type=dict, default_factory=dict
    ),
    Option(
        name='ytdl_enable_whitelist', env_var='YTDL_ENABLE_WHITELIST',
        config_key=('ytdl_opts', 'enable_whitelist'), type=bool,
        default=True
    ),
    Option(
        name='ytdl_whitelist', env_var='YTDL_WHITELIST',
        config_key=('ytdl_opts', 'whitelist'), type=list,
        # TODO: add support for match_filter (requires a custom_opt).
        value_converter=frozenset, default=frozenset((
            'age_limit', 'allsubtitles', 'ap_mso', 'ap_password',
            'ap_username', 'buffersize', 'continuedl', 'default_search',
            'fixup', 'force_generic_extractor', 'format', 'geo_bypass',
            'geo_bypass_country', 'geo_bypass_ip_block', 'hls_use_mpegts',
            'http_chunk_size', 'ignoreerrors', 'include_ads', 'keepvideo',
            'matchtitle', 'max_filesize', 'max_sleep_interval',
            'max_views', 'merge_output_format', 'min_filesize',
            'min_views', 'nooverwrites', 'nopart', 'noplaylist',
            'noresizebuffer', 'outtmpl', 'outtmpl_na_placeholder',
            'password', 'playlistend', 'playlist_items', 'playlistrandom',
            'playlistreverse', 'playliststart', 'ratelimit', 'rejecttitle',
            'restrictfilenames', 'retries', 'simulate', 'skip_download',
            'sleep_interval', 'subtitlesformat', 'subtitleslangs',
            'updatetime', 'username', 'videopassword',
            'write_all_thumbnails', 'writeannotations',
            'writeautomaticsub', 'writedescription', 'writeinfojson',
            'writesubtitles', 'writethumbnail', 'xattr_set_filesize',
            'youtube_include_dash_manifest',
            # yt-dlp options
            # TODO: Add support for 'paths'. It allows downloading to
            #       other locations, so will need to be checked
            #       similarily to 'outtmpl'
            'allow_multiple_audio_streams', 'allow_multiple_video_streams',
            'allow_playlist_files', 'break_on_existing', 'break_on_reject',
            'check_formats', 'dynamic_mpd', 'extractor_retries',
            'force_write_download_archive', 'final_ext', 'format_sort',
            'format_sort_force', 'fragment_retries', 'getcomments',
            'hls_split_discontinuity', 'ignore_no_formats_error', 'overwrites',
            'skip_playlist_after_errors', 'sleep_interval_requests',
            'sleep_interval_subtitles', 'throttledratelimit', 'trim_file_name',
            'windowsfilenames', 'writedesktoplink', 'writeurllink',
            'writewebloclink', 'youtube_include_hls_manifest'
        ))
    ),
    Option(
        name='ytdl_whitelist_add', env_var='YTDL_WHITELIST_ADD',
        config_key=('ytdl_opts', 'whitelist_add'), type=list,
        value_converter=frozenset, default=frozenset()
    ),
    Option(
        name='ytdl_whitelist_remove', env_var='YTDL_WHITELIST_REMOVE',
        config_key=('ytdl_opts', 'whitelist_remove'), type=list,
        value_converter=frozenset, default=frozenset()
    ),
    Option(
        name='ytdl_sensitive_opts', env_var='YTDL_SENSITIVE_OPTS',
        config_key=('ytdl_opts', 'sensitive_opts'), type=list,
        value_converter=tuple,
        default=('ap_password', 'password', 'videopassword')
    ),
    Option(
        name='custom_default_opts', config_key=('custom_opts', 'default_opts'),
        type=dict, default_factory=dict
    ),
    Option(
        name='custom_enable_whitelist',
        env_var='YTDL_ENABLE_CUSTOM_WHITELIST',
        config_key=('custom_opts', 'enable_whitelist'), type=bool,
        default=False
    ),
    Option(
        name='custom_whitelist', env_var='YTDL_CUSTOM_WHITELIST',
        config_key=('custom_opts', 'whitelist'), type=list,
        value_converter=frozenset, default=frozenset()
    ),
    Option(
        name='ytdl_module', env_var='YTDL_MODULE', type=str,
        default='youtube_dl'
    ),
    Option(
        name='ytdl_class', env_var='YTDL_CLASS', type=str,
        default='YoutubeDL'
    ),
    Option(
        name='daterange_module', env_var='YTDL_DATERANGE_MODULE', type=str,
        default=None
    ),
    Option(
        name='daterange_class', env_var='YTDL_DATERANGE_CLASS', type=str,
        default='DateRange'
    ),
    Option(
        name='metadata_module', env_var='YTDL_METADATA_MODULE', type=str,
        default=None
    ),
    Option(
        name='metadata_class', env_var='YTDL_METADATA_CLASS', type=str,
        default='MetadataParserPP'
    )
)
"""List of config options that will be loaded in `config.Config`"""

FLASK_OPTIONS: Final = (
    FlaskCeleryOption(
        name='SERVER_NAME', env_var='YTDL_FLASK_SERVER_NAME', type=str
    ),
    FlaskCeleryOption(
        name='APPLICATION_ROOT', env_var='YTDL_FLASK_APPLICATION_ROOT',
        type=str
    ),
    FlaskCeleryOption(
        name='PREFERRED_URL_SCHEME',
        env_var='YTDL_FLASK_PREFERRED_URL_SCHEME', type=str
    ),
    FlaskCeleryOption(
        name='MAX_CONTENT_LENGTH', env_var='YTDL_FLASK_MAX_CONTENT_LENGTH',
        type=int
    ),
    FlaskCeleryOption(
        name='JSON_AS_ASCII', env_var='YTDL_FLASK_JSON_AS_ASCII', type=bool
    ),
    FlaskCeleryOption(
        name='JSON_SORT_KEYS', env_var='YTDL_FLASK_JSON_SORT_KEYS',
        type=bool
    ),
    FlaskCeleryOption(
        name='JSONIFY_PRETTYPRINT_REGULAR',
        env_var='YTDL_FLASK_JSONIFY_PRETTYPRINT_REGULAR', type=bool
    ),
    FlaskCeleryOption(
        name='JSONIFY_MIMETYPE', env_var='YTDL_FLASK_JSONIFY_MIMETYPE',
        type=str
    )
)
"""List of Flask env-var options that will be added to
`config.Config.flask_config`"""

CELERY_OPTIONS: Final = (
    FlaskCeleryOption(
        name='enable_utc', env_var='YTDL_CELERY_ENABLE_UTC', type=bool
    ),
    FlaskCeleryOption(
        name='timezone', env_var='YTDL_CELERY_TIMEZONE', type=str
    ),
    FlaskCeleryOption(
        name='task_compression', env_var='YTDL_CELERY_TASK_COMPRESSION',
        type=str
    ),
    FlaskCeleryOption(
        name='task_protocol', env_var='YTDL_CELERY_TASK_PROTOCOL', type=int
    ),
    FlaskCeleryOption(
        name='task_publish_retry',
        env_var='YTDL_CELERY_TASK_PUBLISH_RETRY', type=bool
    ),
    FlaskCeleryOption(
        name='task_soft_time_limit',
        env_var='YTDL_CELERY_TASK_SOFT_TIME_LIMIT', type=int
    ),
    FlaskCeleryOption(
        name='task_acks_late', env_var='YTDL_CELERY_TASK_ACKS_LATE',
        type=bool
    ),
    FlaskCeleryOption(
        name='task_acks_on_failure_or_timeout',
        env_var='YTDL_CELERY_TASK_ACKS_ON_FAILURE_OR_TIMEOUT', type=bool
    ),
    FlaskCeleryOption(
        name='task_reject_on_worker_lost',
        env_var='YTDL_CELERY_TASK_REJECT_ON_WORKER_LOST', type=bool
    ),
    FlaskCeleryOption(
        name='task_default_rate_limit',
        env_var='YTDL_CELERY_TASK_DEFAULT_RATE_LIMIT', type=int
    ),
    # 'result_backend' is automatically set to the same value as
    # `database_uri` when it's not explicitly given
    FlaskCeleryOption(
        name='result_backend', env_var='YTDL_CELERY_RESULT_BACKEND',
        type=str
    ),
    FlaskCeleryOption(
        name='result_backend_always_retry',
        env_var='YTDL_CELERY_RESULT_BACKEND_ALWAYS_RETRY', type=bool
    ),
    FlaskCeleryOption(
        name='result_backend_max_sleep_between_retries_ms',
        env_var='YTDL_CELERY_RESULT_BACKEND_MAX_SLEEP_BETWEEN_RETRIES_MS',
        type=int
    ),
    FlaskCeleryOption(
        name='result_backend_base_sleep_between_retries_ms',
        env_var='YTDL_CELERY_RESULT_BACKEND_BASE_SLEEP_BETWEEN_RETRIES_MS',
        type=int
    ),
    FlaskCeleryOption(
        name='result_backend_max_retries',
        env_var='YTDL_CELERY_RESULT_BACKEND_MAX_RETRIES', type=int
    ),
    FlaskCeleryOption(
        name='result_compression',
        env_var='YTDL_CELERY_RESULT_COMPRESSION', type=str
    ),
    FlaskCeleryOption(
        name='result_expires', env_var='YTDL_CELERY_RESULT_EXPIRES',
        type=int
    ),
    FlaskCeleryOption(
        name='result_chord_join_timeout',
        env_var='YTDL_CELERY_RESULT_CHORD_JOIN_TIMEOUT', type=float
    ),
    FlaskCeleryOption(
        name='result_chord_retry_interval',
        env_var='YTDL_CELERY_RESULT_CHORD_RETRY_INTERVAL', type=float
    ),
    FlaskCeleryOption(
        name='database_short_lived_sessions',
        env_var='YTDL_CELERY_DATABASE_SHORT_LIVED_SESSIONS', type=bool
    ),
    FlaskCeleryOption(
        name='broker_url', env_var='YTDL_CELERY_BROKER_URL', type=str
    ),
    FlaskCeleryOption(
        name='broker_read_url', env_var='YTDL_CELERY_BROKER_READ_URL',
        type=str
    ),
    FlaskCeleryOption(
        name='broker_write_url', env_var='YTDL_CELERY_BROKER_WRITE_URL',
        type=str
    ),
    FlaskCeleryOption(
        name='broker_failover_strategy',
        env_var='YTDL_CELERY_BROKER_FAILOVER_STRATEGY', type=str
    ),
    FlaskCeleryOption(
        name='broker_heartbeat', env_var='YTDL_CELERY_BROKER_HEARTBEAT',
        type=float
    ),
    FlaskCeleryOption(
        name='broker_heartbeat_checkrate',
        env_var='YTDL_CELERY_BROKER_HEARTBEAT_CHECKRATE', type=float
    ),
    FlaskCeleryOption(
        name='broker_pool_limit', env_var='YTDL_CELERY_BROKER_POOL_LIMIT',
        type=int
    ),
    FlaskCeleryOption(
        name='broker_connection_timeout',
        env_var='YTDL_CELERY_BROKER_CONNECTION_TIMEOUT', type=float
    ),
    FlaskCeleryOption(
        name='broker_connection_retry',
        env_var='YTDL_CELERY_BROKER_CONNECTION_RETRY', type=bool
    ),
    FlaskCeleryOption(
        name='broker_connection_max_retries',
        env_var='YTDL_CELERY_BROKER_CONNECTION_MAX_RETRIES', type=int
    ),
    FlaskCeleryOption(
        name='broker_login_method',
        env_var='YTDL_CELERY_BROKER_LOGIN_METHOD', type=str
    ),
    FlaskCeleryOption(
        name='worker_concurrency',
        env_var='YTDL_CELERY_WORKER_CONCURRENCY', type=int
    ),
    FlaskCeleryOption(
        name='worker_prefetch_multiplier',
        env_var='YTDL_CELERY_WORKER_PREFETCH_MULTIPLIER', type=int
    ),
    FlaskCeleryOption(
        name='worker_lost_wait', env_var='YTDL_CELERY_WORKER_LOST_WAIT',
        type=float
    ),
    FlaskCeleryOption(
        name='worker_max_tasks_per_child',
        env_var='YTDL_CELERY_WORKER_MAX_TASKS_PER_CHILD', type=int
    ),
    FlaskCeleryOption(
        name='worker_max_memory_per_child',
        env_var='YTDL_CELERY_WORKER_MAX_MEMORY_PER_CHILD', type=int
    ),
    FlaskCeleryOption(
        name='worker_disable_rate_limits',
        env_var='YTDL_CELERY_WORKER_DISABLE_RATE_LIMITS', type=bool
    ),
    FlaskCeleryOption(
        name='worker_state_db', env_var='YTDL_CELERY_WORKER_STATE_DB',
        type=str
    ),
    FlaskCeleryOption(
        name='worker_timer_precision',
        env_var='YTDL_CELERY_WORKER_TIMER_PRECISION', type=float
    ),
    FlaskCeleryOption(
        name='worker_enable_remote_control',
        env_var='YTDL_CELERY_WORKER_ENABLE_REMOTE_CONTROL', type=bool
    ),
    FlaskCeleryOption(
        name='worker_proc_alive_timeout',
        env_var='YTDL_CELERY_WORKER_PROC_ALIVE_TIMEOUT', type=float
    ),
    FlaskCeleryOption(
        name='worker_cancel_long_running_tasks_on_connection_loss',
        env_var=(
            'YTDL_CELERY_'
            'WORKER_CANCEL_LONG_RUNNING_TASKS_ON_CONNECTION_LOSS'
        ),
        type=bool
    ),
    FlaskCeleryOption(
        name='worker_send_task_events',
        env_var='YTDL_CELERY_WORKER_SEND_TASK_EVENTS', type=bool
    ),
    FlaskCeleryOption(
        name='task_send_sent_event',
        env_var='YTDL_CELERY_TASK_SEND_SENT_EVENT', type=bool
    ),
    FlaskCeleryOption(
        name='event_queue_ttl',
        env_var='YTDL_CELERY_EVENT_QUEUE_TTL', type=float
    ),
    FlaskCeleryOption(
        name='event_queue_expires',
        env_var='YTDL_CELERY_EVENT_QUEUE_EXPIRES', type=float
    ),
    FlaskCeleryOption(
        name='event_queue_prefix',
        env_var='YTDL_CELERY_EVENT_QUEUE_PREFIX', type=str
    ),
    FlaskCeleryOption(
        name='event_exchange', env_var='YTDL_CELERY_EVENT_EXCHANGE',
        type=str
    ),
    FlaskCeleryOption(
        name='control_queue_ttl', env_var='YTDL_CELERY_CONTROL_QUEUE_TTL',
        type=float
    ),
    FlaskCeleryOption(
        name='control_queue_expires',
        env_var='YTDL_CELERY_CONTROL_QUEUE_EXPIRES', type=float
    ),
    FlaskCeleryOption(
        name='control_exchange', env_var='YTDL_CELERY_CONTROL_EXCHANGE',
        type=str
    ),
    FlaskCeleryOption(
        name='worker_log_color', env_var='YTDL_CELERY_WORKER_LOG_COLOR',
        type=bool
    ),
    FlaskCeleryOption(
        name='worker_log_format', env_var='YTDL_CELERY_WORKER_LOG_FORMAT',
        type=str
    ),
    FlaskCeleryOption(
        name='worker_task_log_format',
        env_var='YTDL_CELERY_WORKER_TASK_LOG_FORMAT', type=str
    ),
    FlaskCeleryOption(
        name='beat_schedule_filename',
        env_var='YTDL_CELERY_BEAT_SCHEDULE_FILENAME', type=str
    ),
    FlaskCeleryOption(
        name='beat_sync_every', env_var='YTDL_CELERY_BEAT_SYNC_EVERY',
        type=int
    ),
    FlaskCeleryOption(
        name='beat_max_loop_interval',
        env_var='YTDL_CELERY_BEAT_MAX_LOOP_INTERVAL', type=int
    )
)
"""List of Celery env-var options that will be added to
`config.Config.celery_config`"""



"""Flask blueprint for converting exceptions to JSON"""

from __future__ import annotations

__all__ = ('bp', 'handle_exception', 'JSONEncoder')

from typing import TYPE_CHECKING

import flask
import werkzeug

from .error import http_error

if TYPE_CHECKING:
    from typing import Any

bp = flask.Blueprint('json', __name__)


class JSONEncoder(flask.json.JSONEncoder):
    """JSONEncoder that converts unserializable objects to a string

    The object is converted to a str via `repr()`, and a warning is
    also logged
    """
    def default(self, obj: Any) -> Any:
        try:
            # Attempt to convert the object using the additional methods
            # defined by flask.json.JSONEncoder
            return super().default(obj)
        except TypeError:
            flask.current_app.logger.warning(
                'Encountered unserializable object when dumping to json: %r',
                obj
            )
            return repr(obj)


@bp.app_errorhandler(werkzeug.exceptions.HTTPException)
def handle_exception(e: werkzeug.exceptions.HTTPException) -> tuple[
    werkzeug.sansio.response.Response, int
]:
    """Return JSON instead of HTML for HTTP errors"""
    # Start with the correct headers and status code from the error
    response = e.get_response()

    # Create a JSONified response
    #
    # `flask.jsonify` is used instead of calling `flask.json.dumps`
    # directly because it pretty-prints the JSON depending on the app
    # config
    error = http_error.UnspecifiedError(e.name, e.code, e.description)
    jsonified_response = flask.jsonify(error.to_dict())

    # Replace the data in the original response with the JSONified data
    #
    # Mypy complains that `response` doesn't have the attribute `data`.
    # Not sure why, since it's an assignment
    response.data = jsonified_response.data  # type: ignore[attr-defined]
    response.content_type = jsonified_response.content_type

    return response, error.code
